/*  File: gbwt.c
 *  Author: Richard Durbin (rd@sanger.ac.uk)
 *  Copyright (C) Genome Research Limited, 2013
 *-------------------------------------------------------------------
 * Description: genotype suffix-array/BWT based analysis code
 * Exported functions:
 * HISTORY:
 * Last edited: Mar 12 02:57 2013 (rd)
 * Created: Sun Jan  6 12:47:35 2013 (rd)
 *-------------------------------------------------------------------
 */

#include "utils.h"

typedef unsigned char uchar ;

typedef struct SiteStruct {
  int x ;
  int f ;			/* number of 1s, < M */
} Site ;

typedef struct GBWTstruct {
  int N ;
  int M ;
  Array sites ;			/* array of Site */
  int *a ;
  uchar *y ;
  Array yz ;			/* compressed BWT array of uchar */
} GBWT ;

static int isCheck = 0 ;
static int isStats = 0 ;

/***************** rusage for timing information ******************/

#include <sys/resource.h>
#define RUSAGE_SELF 0

#ifdef RUSAGE_STRUCTURE_DEFINITIONS

struct rusage {
  struct timeval ru_utime; /* user time used */
  struct timeval ru_stime; /* system time used */
  long ru_maxrss;          /* integral max resident set size */
  long ru_ixrss;           /* integral shared text memory size */
  long ru_idrss;           /* integral unshared data size */
  long ru_isrss;           /* integral unshared stack size */
  long ru_minflt;          /* page reclaims */
  long ru_majflt;          /* page faults */
  long ru_nswap;           /* swaps */
  long ru_inblock;         /* block input operations */
  long ru_oublock;         /* block output operations */
  long ru_msgsnd;          /* messages sent */
  long ru_msgrcv;          /* messages received */
  long ru_nsignals;        /* signals received */
  long ru_nvcsw;           /* voluntary context switches */
  long ru_nivcsw;          /* involuntary context switches */
};

struct timeval {
  time_t       tv_sec;   /* seconds since Jan. 1, 1970 */
  suseconds_t  tv_usec;  /* and microseconds */
} ;

#endif /* RUSAGE STRUCTURE_DEFINITIONS */

void timeUpdate (void)
{
  static int isFirst = 1 ;
  static struct rusage rOld ;
  struct rusage rNew ;
  int secs, usecs ;

  getrusage (RUSAGE_SELF, &rNew) ;
  if (!isFirst)
    { secs = rNew.ru_utime.tv_sec - rOld.ru_utime.tv_sec ;
      usecs =  rNew.ru_utime.tv_usec - rOld.ru_utime.tv_usec ;
      if (usecs < 0) { usecs += 1000000 ; secs -= 1 ; }
      fprintf (stderr, "user\t%d.%06d", secs, usecs) ;
      secs = rNew.ru_stime.tv_sec - rOld.ru_stime.tv_sec ;
      usecs =  rNew.ru_stime.tv_usec - rOld.ru_stime.tv_usec ;
      if (usecs < 0) { usecs += 1000000 ; secs -= 1 ; }
      fprintf (stderr, "\tsystem\t%d.%06d", secs, usecs) ;
      fprintf (stderr, "\tmax_RSS\t%ld", rNew.ru_maxrss - rOld.ru_maxrss) ;
      fputc ('\n', stderr) ;
    }
  else
    isFirst = 0 ;
  rOld = rNew ;
}

/****************** Y compression and decompression *************************/
/****************** low level - does not know about GBWT structures *********/

/* yPack3 is a three level run length encoding: n times value
   yp & 0x80 = value
   yp & 0x40 == 0 implies n = yp & 0x3f
   yp & 0x40 == 1 implies
     yp & 0x20 == 0 implies n = (yp & 0x1f) << 6
     yp & 0x20 == 1 implies n = (yp & 0x1f) << 11
   This allows coding runs of length up to 64 * 32 * 32 = 64k in 3 bytes.
   Potential factor of ~1000 over a bit array.
   Build a lookup to avoid conditional operation in uncompression.
*/

#define Y_SENTINEL 2
static int yp3decode[128] ;
#define ENCODE_MAX1 64		     /* ~64 */
#define ENCODE_MAX2 ((95-63) << 6)   /* ~1k - is this 32 or 31?*/
#define ENCODE_MAX3 ((127-96) << 11) /* ~64k - ditto */

void yPack3init (void)
{
  int n ;
  for (n = 0 ; n < 64 ; ++n) yp3decode[n] = n ;
  for (n = 64 ; n < 96 ; ++n) yp3decode[n] = (n-64) << 6 ;
  for (n = 96 ; n < 128 ; ++n) yp3decode[n] = (n-96) << 11 ;
}

inline int yPack3Add (uchar yy, uchar *yzp, int n)
{ 
  uchar *yzp0 = yzp ;

  yy <<= 7 ;			/* first move the actual symbol to the top bit */

  while (n >= ENCODE_MAX3) { *yzp++ = yy | 0x7f ; n -= ENCODE_MAX3 ; }
  if (n >= ENCODE_MAX2) { *yzp++ = yy | 0x60 | (n >> 11) ; n &= 0x7ff ; }
  if (n >= ENCODE_MAX1) { *yzp++ = yy | 0x40 | (n >> 6) ; n &= 0x3f ; }
  if (n) { *yzp++ = yy | n ; }
  return yzp - yzp0 ;
}

int yPack3 (uchar *yp, int M, uchar *yzp) /* pack M chars from y into yz - return number of chars added */
{ 
  int m = 0, m0 ;
  uchar *yzp0 = yzp ;
  uchar ym ;
  
  while (m < M)			/* NB this relies on the M'th character not being 0 or 1 */
    { ym = *yp++ ; m0 = m++ ;
      while (*yp == ym) { ++m ; ++yp ; }
      yzp += yPack3Add (ym, yzp, m-m0) ;
    }
  return yzp - yzp0 ;
}

int yUnpack3 (uchar *yzp, int M, uchar *yp, int *n1) /* unpack yz into M chars in y and move along - return number of chars unpacked - n1 is number of 1s */
{
  int m = 0 ;
  uchar *yzp0 = yzp ;
  uchar yz ;
  int n ;

  if (n1) *n1 = 0 ;
  while (m < M)
    { yz = *yzp++ ;
      n = yp3decode[yz & 0x7f] ;
      m += n ;
      yz = yz >> 7 ;
      if (yz && n1) *n1 += n ;
      if (n > 63)
	{ memset (yp, yz, n) ;
	  yp += n ;
	}
      else
	while (n--) *yp++ = yz ;
    }

  return yzp - yzp0 ;
}

 uchar *yUnpack3interval (uchar *yy, int min, int max)
 { 
   /* idea here is that we can just unpack part of y */
   /* actually, I can update a and d and u in chunks... */
 }

/************ key gbwt extension algorithm ************/

void initialiseA (int *a, int M)
{
  int i ;
  for (i = 0 ; i < M ; ++i) *a++ = i ;
}

void updateA (int *a, uchar *y, int M)		/* algorithm 1 in the manuscript */
{
  static int *b = 0 ;
  int u = 0 ;			/* calculate v = i-u */
  int i ;

  if (!b)			/* initialise */
    b = myalloc (M * sizeof(int)) ;
    
  for (i = 0 ; i < M ; ++i)
    if (*y++ == 0)
      a[u++] = a[i] ;
    else			/* y[i] == 1, since bi-allelic */
      b[i-u] = a[i] ;

  memcpy (a+u, b, (M-u)*sizeof(int)) ;
}

/************** core gbwt options *********************/

GBWT *gbwtCreate (int M)
{
  GBWT *g = (GBWT*) mycalloc (1, sizeof (GBWT)) ; /* cleared so elements default to 0 */

  g->M = M ;
  g->a = (int*) myalloc (M*sizeof(int)) ;
  g->y = (uchar*) myalloc (M+1) ;
  g->y[M] = Y_SENTINEL ;

  return g ;
}

void gbwtDestroy (GBWT *g)
{
  if (g->a) free (g->a) ;
  if (g->y) free (g->y) ;
  if (g->sites) arrayDestroy (g->sites) ;
  if (g->yz) arrayDestroy (g->yz) ;
  free (g) ;
}

void gbwtWrite (GBWT *g, FILE *fp)
{
  int n ;

  if (!g || !g->yz) die ("gbwtWrite called without a valid gbwt") ;

  n = arrayMax(g->yz) ;
  if (fwrite ("GBWT", 1, 4, fp) != 4)
    die ("error writing GBWT in gbwtWrite") ;
  if (fwrite (&g->M, sizeof(int), 1, fp) != 1)
    die ("error writing n in gbwtWrite") ;
  if (fwrite (&g->N, sizeof(int), 1, fp) != 1)
    die ("error writing n in gbwtWrite") ;
  if (fwrite (&n, sizeof(int), 1, fp) != 1)
    die ("error writing n in gbwtWrite") ;
  if (fwrite (arrp(g->yz, 0, uchar), sizeof(uchar), arrayMax(g->yz), fp) != arrayMax(g->yz))
    die ("error writing data in gbwtWrite") ;

  fprintf (stderr, "written %d chars gbwt: M, N are %d, %d\n", arrayMax(g->yz), g->M, g->N) ;
}

void gbwtWriteSites (GBWT *g, FILE *fp)
{
  int i ;

  if (!g || !g->sites) die ("gbwtWriteSites called without a valid gbwt") ;

  for (i = 0 ; i < g->N ; ++i)
    fprintf (fp, "%d\n", arrp(g->sites, i, Site)->x) ;
  if (ferror (fp)) die ("error writing sites file") ;

  fprintf (stderr, "written %d sites from %d to %d\n", g->N, 
	   arrp(g->sites, 0, Site)->x, arrp(g->sites, g->N-1, Site)->x) ;
}

GBWT *gbwtRead (FILE *fp) 
{
  int m, n ;
  GBWT *g ;
  static char tag[5] = "test" ;

#ifdef TRANSITION		/* only used when moving from old to new file format */
  m = 101000 ;
  n = 370264 ;
#else
  if (fread (tag, 1, 4, fp) != 4 || strcmp (tag, "GBWT"))
    die ("error reading GBWT in gbwtRead") ;
  if (fread (&m, sizeof(int), 1, fp) != 1)
    die ("error reading n in gbwtRead") ;
  if (fread (&n, sizeof(int), 1, fp) != 1)
    die ("error reading n in gbwtRead") ;
#endif
  g = gbwtCreate (m) ;
  g->N = n ;
  if (fread (&n, sizeof(int), 1, fp) != 1)
    die ("error reading gbwt file") ;
  g->yz = arrayCreate (n, uchar) ;
  array(g->yz, n-1, uchar) = 0 ; /* sets arrayMax */
  if (fread (arrp(g->yz, 0, uchar), sizeof(uchar), n, fp) != n)
    die ("error reading data in gbwt file") ;

  fprintf (stderr, "read gbwt file: M, N are %d, %d\n", g->M, g->N) ;
  return g ;
}

void gbwtReadSites (GBWT *g, FILE *fp)
{
  int n = 0 ;
  if (!g) die ("gbwtReadSites called without a valid gbwt") ;

  g->sites = arrayReCreate (g->sites, g->N, Site) ;
  while (!feof(fp))
    fscanf (fp, "%d\n", &(arrayp(g->sites, n++, Site)->x)) ;
  if (ferror (fp)) die ("error reading sites file") ;

  if (n != g->N)
    { fprintf (stderr, "sites file contains %d sites not %d as in gbwt - reject these sites\n",
	       n, g->N) ;
      arrayDestroy (g->sites) ; g->sites = 0 ;
    }
  else
    fprintf (stderr, "read %d sites from file\n", n) ;
}

GBWT *gbwtSubSample (GBWT *gOld, int start, int m)
{
  GBWT *gNew = gbwtCreate (m) ;
  int i, j, nOld = 0 , nNew, k ;
  uchar *x, *yz ;

  if (!gOld || !gOld->yz) die ("subsample called without an valid existing gbwt") ;
  gNew->N = gOld->N ;

  x = (uchar*) myalloc (gNew->M) ;
  yz = (uchar*) myalloc (gNew->M) ;

  initialiseA (gOld->a, gOld->M) ;
  initialiseA (gNew->a, gNew->M) ;
  gNew->yz = arrayCreate (gNew->N*8, uchar) ;
  for (i = 0 ; i < gOld->N ; ++i)
    { nOld += yUnpack3 (arrp(gOld->yz,nOld,uchar), gOld->M, gOld->y, 0) ;
      for (j = 0 ; j < gOld->M ; ++j)
	if ((k = gOld->a[j] - start) >= 0 && k < m)
	  x[k] = gOld->y[j] ;
      for (j = 0 ; j < gNew->M ; ++j)
	gNew->y[j] = x[gNew->a[j]] ;
      nNew = yPack3 (gNew->y, gNew->M, yz) ;
      for (j = 0 ; j < nNew ; ++j) array(gNew->yz,arrayMax(gNew->yz),uchar) = yz[j] ;
      updateA (gOld->a, gOld->y, gOld->M) ;
      updateA (gNew->a, gNew->y, gNew->M) ;
    }

  gNew->sites = gOld->sites ; gOld->sites = 0 ;
  gbwtDestroy (gOld) ;
  free(x) ; free(yz) ;
  return gNew ;
}

GBWT *gbwtSubSites (GBWT *gOld, double fmin, double frac)
{
  int M = gOld->M ;
  GBWT *gNew = gbwtCreate (M) ;
  int i, j, nOld = 0, nNew, k, n1, thresh = M*fmin  ;
  double bit = 0.0 ;
  uchar *x, *yz ;

  if (!gOld || !gOld->sites) die ("subsites without an existing gbwt") ;

  x = (uchar*) myalloc (M) ;
  yz = (uchar*) myalloc (M) ;
  gNew->sites = arrayCreate (4096, Site) ;

  initialiseA (gOld->a, M) ;
  initialiseA (gNew->a, M) ;
  gNew->yz = arrayCreate (gNew->N*8, uchar) ;
  for (i = 0 ; i < gOld->N ; ++i)
    { nOld += yUnpack3 (arrp(gOld->yz,nOld,uchar), M, gOld->y, &n1) ;
      if ((n1 > thresh) && ((bit += frac) > 1.0))
	{ for (j = 0 ; j < M ; ++j)
	    x[gOld->a[j]] = gOld->y[j] ;
	  for (j = 0 ; j < M ; ++j)
	    gNew->y[j] = x[gNew->a[j]] ;
	  nNew = yPack3 (gNew->y, M, yz) ;
	  for (j = 0 ; j < nNew ; ++j) array(gNew->yz,arrayMax(gNew->yz),uchar) = yz[j] ;
	  updateA (gNew->a, gNew->y, M) ;
	  array(gNew->sites, gNew->N++, Site) = arr(gOld->sites, i, Site)  ;
	  bit -= 1.0 ;
	}  
      updateA (gOld->a, gOld->y, M) ;
    }

  fprintf (stderr, "subsites with fmin %f, frac %f leaves %d sites\n", fmin, frac, gNew->N) ;

  gbwtDestroy (gOld) ;
  free(x) ; free(yz) ;
  return gNew ;
}

/****************** MaCS file parsing *****************/

void parseMacsHeader (FILE *fp, int *M, double *L)	/* parse MACS file header */
{
  if (strcmp (fgetword (fp), "COMMAND:")) die ("MaCS COMMAND line not found") ;
  fgetword (fp) ; 		/* the command */
  *M = atoi(fgetword(fp)) ; if (!*M) die ("failed to get M") ;
  *L = atof(fgetword(fp)) ; if (!*L) die ("failed to get L") ;
  while (fgetc(fp) != '\n') ;	/* ignore rest of line */
  
  if (strcmp (fgetword (fp), "SEED:")) die ("SEED line not found") ;
  while (fgetc(fp) != '\n') ;	/* ignore rest of line */
}

BOOL parseMacsSite (FILE *fp, Site *s, int M, double L, uchar *yp) /* parse MACS site line */
{
  static uchar conv[256] ;
  static BOOL isFirst = TRUE ;
  int number ;

  if (isFirst) { conv['0'] = 0 ; conv['1'] = 1 ; isFirst = FALSE ; }
  if (feof (fp)) return FALSE ;

  if (strcmp (fgetword (fp), "SITE:")) 
    { if (feof (fp)) return FALSE ;
      die ("MaCS SITE line not found") ;
    }
  number = atoi(fgetword(fp)) ;	/* this is the site number */
  s->x = (int) (L * atof(fgetword(fp))) ;
  atof(fgetword(fp)) ;		/* ignore the time */
  while (M--)
    *yp++ = conv[getc(fp)] ;
  if (feof (fp)) return FALSE ;
  if (getc(fp) != '\n') die ("end of line error for MaCS SITE %d", number) ;

  return TRUE ;
}

GBWT *readMacs (FILE *fp)
{
  GBWT *g ;
  int M ;
  double L ;
  int i,j, nxPack, nyPack, nUnpack, n1 ;
  int nxTot = 0, nyTot = 0 ;
  uchar *x, *yz, *y2 ;		/* original, sorted, compressed, uncompressed */
  int *a ;

  parseMacsHeader (fp, &M, &L) ;
  g = gbwtCreate (M) ;
  x = (uchar*) myalloc (M+1) ; x[M] = Y_SENTINEL ;	/* sentinel required for packing */
  yz = (uchar*) myalloc (M) ;
  y2 = (uchar*) myalloc (M) ;

  g->M = M ;
  g->sites = arrayCreate(4096, Site) ;
  g->yz = arrayCreate(4096*32, uchar) ;
  g->N = 0 ;
  initialiseA (g->a, M) ;
  while (parseMacsSite (fp, arrayp(g->sites,g->N,Site), M, L, x))
    { for (j = 0 ; j < M ; ++j) g->y[j] = x[g->a[j]] ; /* next character in sort order: BWT */
      nyPack = yPack3 (g->y, M, yz) ;
      for (j = 0 ; j < nyPack ; ++j) array(g->yz,arrayMax(g->yz),uchar) = yz[j] ;
      updateA (g->a, g->y, M) ;
      if (isCheck)
	{ int nx1 ;
	  nUnpack = yUnpack3 (yz, M, y2, &n1) ;
	  for (j = 0 ; j < M ; j++)
	    if (y2[j] != g->y[j])
	      die ("Mismatch y2:%d != y:%d at %d line %d", y2[j], g->y[j], j, i) ;
	  if (nUnpack != nyPack)
	    die ("Unpack %d is not equal to pack %d line %d", nUnpack, nyPack, i) ;
	  nx1 = 0 ; for (j = 0 ; j < M ; ++j) if (x[j]) ++nx1 ;
	  if (nx1 != n1)
	    die ("Mismatch nx1:%d != n1:%d line %d", nx1, n1) ;
      	}
      if (isStats)
	{ if (!isCheck) nUnpack = yUnpack3 (yz, M, y2, &n1) ;
	  nxPack = yPack3 (x, M, yz) ;
	  printf ("Site\t%d\t%d\t%d\t%d\t%d\n", 
		  g->N, arrp(g->sites,g->N,Site)->x, n1, nxPack, nyPack) ;
	  nxTot += nxPack ;
	  nyTot += nyPack ;
      	}
      g->N++ ;
    }

  fprintf (stderr, "read MaCS file: M, N are\t%d\t%d\n", M, g->N) ;
  if (isStats)
    fprintf (stderr, "                xtot, ytot are\t%d\t%d\n", nxTot, nyTot) ;

  free(x) ; free (yz) ; free (y2) ;

  return g ;
}

/*************** write haplotypes ******************/

void writeHaplotypes (FILE *fp, GBWT *g)
{
  int M = g->M ;
  int i, j, n = 0 ;
  int *a ;
  uchar *x ;

  x = (uchar*) myalloc (M) ;
  initialiseA (g->a, M) ;
  for (i = 0 ; i < g->N ; ++i)
    { n += yUnpack3 (arrp(g->yz,n,uchar), M, g->y, 0) ;
      for (j = 0 ; j < M ; ++j) x[g->a[j]] = g->y[j] ;
      for (j = 0 ; j < M ; ++j) putc (x[j]?'1':'0', fp) ;
      putc ('\n', fp) ;
      updateA (g->a, g->y, M) ;
    }

  fprintf (stderr, "written haplotype file: %d rows of %d\n", g->N, g->M) ;
}

/**************** now matching a new sequence **********************/
/*******************************************************************/

void updateA_D_U (int *a, int *d, uchar *y, int M, int k,	/* update a[], d[], y[], u[] */
		  int *aNew, int *dNew, int *uNew)
{
  static int *b = 0, *e ;
  int i, u = 0 ;		/* calculate v = i-u */
  int p = k+1 ; int q = k+1 ;

  if (!b)			/* initialise */
    { b = myalloc (M * sizeof(int)) ;
      e = myalloc (M * sizeof(int)) ;
    }
    
  for (i = 0 ; i < M ; ++i)
    { if (d[i] > p) p = d[i] ;
      if (d[i] > q) q = d[i] ;
      if (*y++ == 0)		/* NB x[a[i]] = y[i] in manuscript */
	{ *aNew++ = a[i] ;
	  *dNew++ = p ;
	  u++ ; p = 0 ;
	}
      else			/* y[i] == 1, since bi-allelic */
	{ b[i-u] = a[i] ;
	  e[i-u] = q ;
	  q = 0 ;
	}
      *uNew++ = u ;
    }
  memcpy (aNew, b, (M-u)*sizeof(int)) ;
  memcpy (dNew, e, (M-u)*sizeof(int)) ;
}

void gbwtTest (GBWT *g, FILE *fp)
{
  GBWT *gTest = gbwtRead (fp) ;
  int i, j, k, n = 0 ;
  int **a, **d, **u, *c ;
  uchar **y, **z ;
  int f,g,e,f1,g1,e1 ;

  if (gTest->N == g->N)
    { a = (int**) myalloc ((g->N+1)*sizeof(int*)) ;
      d = (int**) myalloc ((g->N+1)*sizeof(int*)) ;
      u = (int**) myalloc (g->N*sizeof(int*)) ;
      y = (uchar**) myalloc (g->N*sizeof(uchar*)) ;
      c = (int*) myalloc (g->N*sizeof(int)) ;
      a[0] = (int*) myalloc (g->M*sizeof(int)) ;
      initialiseA (a[0], g->M) ;
      d[0] = (int*) mycalloc (g->M,sizeof(int)) ;
      for (i = 0 ; i < g->N ; ++i)
	{ a[i+1] = (int*) myalloc (g->M*sizeof(int)) ;
	  d[i+1] = (int*) myalloc (g->M*sizeof(int)) ;
	  u[i] = (int*) myalloc (g->M*sizeof(int)) ;
	  y[i] = (uchar*) myalloc (g->M*sizeof(uchar)) ;
	  n += yUnpack3 (arrp(g->yz,n,uchar), g->M, y[i], &c[i]) ;
	  updateA_D_U (a[i], d[i], y[i], g->M, i, a[i+1], d[i+1], u[i]) ;
	}
      fprintf (stderr, "built data structures for test\n") ;
      timeUpdate () ;

      z = (uchar**) myalloc (gTest->M*sizeof(uchar*)) ;
      for (j = 0 ; j < gTest->M ; ++j) z[j] = (uchar*) myalloc (gTest->N*sizeof(uchar)) ;
      initialiseA (gTest->a, gTest->M) ;
      n = 0 ;
      for (i = 0 ; i < gTest->N ; ++i)
	{ n += yUnpack3 (arrp(gTest->yz,n,uchar), gTest->M, gTest->y, 0) ;
	  updateA (gTest->a, gTest->y, gTest->M) ;
	  for (j = 0 ; j < gTest->M ; ++j) z[gTest->a[j]][i] = gTest->y[j] ;	  
	}
      fprintf (stderr, "extracted test sequences\n") ;
      timeUpdate () ;

#ifdef OUT_OF_TIME
      for (j = 0 ; j < gTest->M ; ++j) /* now test each test sequence in turn */
	{ uchar *zz = z[j] ;
	  e = 0 ; f = 0 ; g = gTest->M ;
	  for (k = 0 ; k < g->N ; ++k)
	    { if (z[k] == 0) { f1 = u[k][f] ; g1 = u[k][g] ; }
	      else { f1 = c[k] + f - u[k][f] ; g1 = c[k] + g - u[k][g] ; }
	      if (f1 == g1)
		{ for (i = f ; i < g ; ++i) /* report matches! */
		    { printf ("match test %d to ref %d [%d,%d)\n", j, a[k][i], e, k) ;
		      matchTot++ ;
		    }
		  e1 = d[k+1][f1] - 1 ; /* now find new e1 */
		  if (zz[e1] == 0 && f1 > 0)
		    { f1 = g1 - 1 ; 
		      while (zz[e1-1] == !!!!!!! ) do e1 = e1 - 1 ; 
		    }
		  else
		}
	      e = e1 ; f = f1 ; g = g1 ;
	    }
	}
#endif
    }
  else
    fprintf (stderr, "test gbwt has different length %d from main gbwt %d - no test done",
	     gTest->N, g->N) ;
  
  gbwtDestroy (gTest) ;
}

/*********************************************************/

/* a couple of utilities */
#define FOPEN(name,mode)  if (!strcmp (argv[1], "-")) fp = stdin ; else if (!(fp = fopen (argv[1],mode))) die ("failed to open macs file %s", name, argv[1])
#define FCLOSE if (strcmp(argv[1], "-")) fclose(fp)

int main (int argc, char *argv[])
{
  FILE *fp ;
  GBWT *g = 0 ;
  Array test ;

  yPack3init () ;

  --argc ; ++argv ;
  if (!argc)			/* print help */
    { fprintf (stderr, "Usage: gbwt -<command> [options] ...\n") ;
      fprintf (stderr, "Commands: -check            do various checks\n") ;
      fprintf (stderr, "          -stats            print stats depending on commands; writes to stdout\n") ;
      fprintf (stderr, "          -macs <file>      read MaCS output file; '-' for stdin\n") ;
      fprintf (stderr, "          -haps <file>      write haplotype file; '-' for stdout\n") ;
      fprintf (stderr, "          -write <file>     write gbwt file; '-' for stdout\n") ;
      fprintf (stderr, "          -writeSites <file>  write sites file; '-' for stdout\n") ;
      fprintf (stderr, "          -read <file>     read gbwt file; '-' for stdin\n") ;
      fprintf (stderr, "          -readSites <file>   read sites file; '-' for stdin\n") ;
      fprintf (stderr, "          -subsites <fmin> <frac>  subsample <frac> sites with AF > <fmin>\n") ;
      fprintf (stderr, "          -subsample <start> <n>  subsample <n> samples from index <start>\n") ;
      fprintf (stderr, "          -test <file>      read samples in gbwt test file\n") ;
    }

  timeUpdate() ;
  while (argc) {
    if (!(**argv == '-'))
      die ("not well formed command %s\nType gbwt without arguments for help", *argv) ;
    else if (!strcmp (argv[0], "-check"))
      { isCheck = 1 ;
	argc -= 1 ; argv += 1 ;
      }
    else if (!strcmp (argv[0], "-stats"))
      { isStats = 1 ;
	argc -= 1 ; argv += 1 ;
      }
    else if (!strcmp (argv[0], "-macs") && argc > 1)
      { if (g) gbwtDestroy (g) ;
	FOPEN("macs","r") ; g = readMacs (fp) ;	FCLOSE ;
	argc -= 2 ; argv += 2 ;
      }
    else if (!strcmp (argv[0], "-haps") && argc > 1)
      { FOPEN("haps","w") ; writeHaplotypes (fp, g) ; FCLOSE ;
	argc -= 2 ; argv += 2 ;
      }
    else if (!strcmp (argv[0], "-read") && argc > 1)
      { if (g) gbwtDestroy (g) ;
	FOPEN("read","r") ; g = gbwtRead (fp) ; FCLOSE ;
	argc -= 2 ; argv += 2 ;
      }
    else if (!strcmp (argv[0], "-readSites") && argc > 1)
      { FOPEN("readSites","r") ; gbwtReadSites (g, fp) ; FCLOSE ;
	argc -= 2 ; argv += 2 ;
      }
    else if (!strcmp (argv[0], "-write") && argc > 1)
      { FOPEN("write","w") ; gbwtWrite (g, fp) ; FCLOSE ;
	argc -= 2 ; argv += 2 ;
      }
    else if (!strcmp (argv[0], "-writeSites") && argc > 1)
      { FOPEN("writeSites","w") ; gbwtWriteSites (g, fp) ; FCLOSE ;
	argc -= 2 ; argv += 2 ;
      }
    else if (!strcmp (argv[0], "-test") && argc > 1)
      { FOPEN("test","r") ; gbwtTest (g, fp) ; FCLOSE ;
	argc -= 2 ; argv += 2 ;
      }
    else if (!strcmp (argv[0], "-subsample") && argc > 2)
      { int start = atoi (argv[1]) ;
	int n = atoi (argv[2]) ;
	if (start >= 0 && n > 0)
	  g = gbwtSubSample (g, start, n) ;
	else
	  fprintf (stderr, "start %d, n %d arguments for subsample must be >0\n", start, n) ;
	argc -= 3 ; argv += 3 ;
      }
    else if (!strcmp (argv[0], "-subsites") && argc > 2)
      { double freqmin = atof (argv[1]) ;
	double fraction = atof (argv[2]) ;
	if (freqmin >= 0 && freqmin < 1 && fraction > 0 && fraction < 1)
	  g = gbwtSubSites (g, freqmin, fraction) ;
	else
	  fprintf (stderr, "fmin %f, frac %f arguments for subsites must be between 0 and 1\n", 
		   freqmin, fraction) ;
	argc -= 3 ; argv += 3 ;
      }
    else
      die ("unrecognised command %s\nType gbwt without arguments for help", *argv) ;
    timeUpdate() ;
  }
}

/*************************************/
